module GiJob
  module Models
    module Concerns
      module GiJobTransactionable

        extend ActiveSupport::Concern

        included do # begin included

          serialize  :parameter
          serialize  :process_progress

          belongs_to :owner, polymorphic: true, optional: true
          has_many   :gi_job_logs,   dependent: :destroy
          has_many   :gi_job_files,  dependent: :destroy

          enum status: {
            registerd:  100,
            running:    200,
            completed:  300,
            error:      400,
          }, _prefix: true

          enum process_status: {
            none:       100,
            suspended:  200,
            terminated: 300,
          }, _prefix: true

          enum command: {
            none:       100,
            stop:       200,
          }, _prefix: true


          scope :search_jobs, -> (params) {
            search_by_id(params[:id])
              .search_by_name(params[:name])
              .search_by_status(params[:status])
              .order_by(params[:order_by])
          }

          scope :search_by_id, -> (value) {
            where(id: value) if value.present?
          }

          scope :search_by_name, -> (value) {
            where(name: value) if value.present?
          }

          scope :search_by_status, -> (value) {
            v = value.is_a?(String) ? GiJobTransaction.statuses[value] : value
            where(status: v) if v.present?
          }

          scope :order_by, -> (order_by = nil){
            if order_by.present?
              order("#{order_by}")
            else
              order("gi_job_transactions.updated_at desc")
            end
          }

          scope :ghosts, -> (no_reaction_minutes) {
            # TODO 更新が無いjobを検出する
            # 本来であればちゃんとprocessキルがされる際にstatusを更新する
            # これは、何らかの原因で jobレコードが更新さずに jobプロセスが死んでしまった場合の暫定的な救済処置である

            to_datetime = Time.zone.now - no_reaction_minutes.minutes
            where(process_status: :none)
              .where(status: [GiJobTransaction.statuses[:registerd], GiJobTransaction.statuses[:running]])
              .where("gi_job_transactions.updated_at < ?", to_datetime)
          }

          scope :suspended, -> () {
            where(process_status: :suspended)
          }

          scope :ended, -> () {
            where(status: [GiJobTransaction.statuses[:completed], GiJobTransaction.statuses[:error]])
          }

          scope :old_ended, -> (time_limit = 14.days) {
            # 既に終了して2週間起つjobを検出する
            time_now = Time.now
            to_datetime =  time_now - time_limit
            Rails.logger.ap("delete_old_file #{time_now} - #{time_limit} = to_datetime: #{to_datetime}")
            where(process_status: :none)
              .ended
              .where("gi_job_transactions.updated_at < ?", to_datetime)
          }

          scope :has_job_file_data_sources, -> () {
            # 有効なdata_sourcesを持っているジョブを検出する
            eager_load(:gi_job_files)
              .where("gi_job_files.file is not null")
          }

          scope :past_ended_same_type_jobs, -> (gi_job_transaction) {
            where(
              owner: gi_job_transaction.owner,
              name: gi_job_transaction.name
            ).where(
              "gi_job_transactions.created_at < ?", gi_job_transaction.created_at
            ).ended.where.not(
              id: gi_job_transaction.id
            )
          }

        end # end included

        class_methods do # begin class_methods

          def delete_old_for_one_shot(gi_job_transaction)
            # 単発ジョブ向け。過去のjob_transactionを削除する。
            self.past_ended_same_type_jobs(gi_job_transaction).destroy_all
          end

        end # end class_methods

        def append_log_info(**args) append_log(level: :info, **args) end
        def append_log_warning(**args) append_log(level: :warning, **args) end
        def append_log_error(**args) append_log(level: :error, **args) end

        def append_log(level: :info, position: nil, division: :none, description: nil)
          GiJobLog.create!({
            gi_job_transaction: self,
            level: GiJobLog.levels[level],
            position: position,
            division: division,
            description: description
          })
        end

        def append_log(level: :info, position: nil, division: :none, description: nil)
          GiJobLog.create!({
            gi_job_transaction: self,
            level: GiJobLog.levels[level],
            position: position,
            division: division,
            description: description
          })
        end

        def create_job_logs_hash
          hash = {
            info: [],
            warning: [],
            error: [],
          }
          self.gi_job_logs.each {|job_log| hash[job_log.level.to_sym] << job_log}
          hash
        end

        def counts_ended
          (counts_completed || 0) + (counts_errord || 0)
        end

        def has_error_log?
          gi_job_logs.any?{|job_log| job_log.error?}
        end

        def to_simple
          job_logs_hash = create_job_logs_hash
          {
            id: self.id,
            name: self.name,
            status: self.status,
            counts_all: self.counts_all,
            counts_progress: self.counts_progress,
            counts_completed: self.counts_completed,
            counts_errord: self.counts_errord,
            process_status: self.process_status,
            created_at: self.created_at,
            updated_at: self.updated_at,
            gi_job_logs: self.gi_job_logs,
            job_logs_hash: job_logs_hash,
            has_error_log: job_logs_hash[:error].present?,
          }
        end

        def status_ended?
          status_completed? || status_error?
        end

        def create_file!(**params)
          GiJobFile.create_file!(gi_job_transaction: self, **params)
        end

        def log_files_filter_common_warns
          gi_job_log_files.filter do |gi_job_log_file|
            gi_job_log_file.file_name.start_with?("警告")
          end
        end

        def log_files_filter_common_errors
          gi_job_log_files.filter do |gi_job_log_file|
            gi_job_log_file.file_name.start_with?("エラー")
          end
        end


        ### for displays
        def to_string_line
          [
            "#{self.id}",
            "#{self.updated_at}",
            "#{self.owner_type}(#{self.owner_id})",
            "#{self.name}",
            "#{self.status}",
            "#{self.counts_progress}/#{self.counts_all}",
            "errord: #{self.counts_errord}",
          ].join(", ")
        end


      end # end GiJobTransactionable
    end
  end
end
